/*
  * This software reads temperature from all four the ADT7420 sensors on I2C.
  *
  * It starts only if ADT7420_lock file does not exist in /tmp directory. This file
  * exists when the program runs and needs to be erased by /etc/rc.local.
  *
  * This binary access /dev/mem. For running from an user account use sticky permissions
  *  and set capabilities.
  *
  * Compilation:
  * gcc -O9 -std=c99 ADT7420_1_all.c -lbcm2835 -o ADT7420_1_all
  *
  * This client is Raspberry Pi specific and uses Mike McCauley's
  * bcm2835 library http://www.airspayce.com/mikem/bcm2835/index.html
  * instead of the I2C drivers in the kernel and was derived form
  * Martin Oldfield's  example (https://github.com/mjoldfield/adt74x0/blob/master/adt74x0b.c)
  * in RADIOSPORT company (Jan.Dvorak@ForYourLab.com).
  *
  * ADT7420 datasheet can be found at:
  *   http://www.analog.com/en/mems-sensors/digital-temperature-sensors/adt7420/products/product.html
  *
  * LICENSE
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of
  * the License, or (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  * General Public License for more details at
  * http://www.gnu.org/copyleft/gpl.html
  *
  *
*/


#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <bcm2835.h>


/* ADT7420 I2C address */
#define ADT7420_ADDR_1  0x48
#define ADT7420_ADDR_2  0x49
#define ADT7420_ADDR_3  0x4A
#define ADT7420_ADDR_4  0x4B
#define ADT7420_CFG  0xC0 // 0x80 for 16bit fast, 0xC0 for 16 bit, 1 SPS

#define LOCK_FILE "/tmp/ADT7420_lock"
#define QUIT_FILE "/tmp/ADT7420_quit"


/* I2C registers in ADT74x0 */
#define T_MSB  0x00
#define T_LSB  0x01
#define STATUS 0x02
#define CONFIG 0x03
#define IDREG  0x0b
#define RESET  0x2f

static int init_adt74x0(const uint8_t addr);
static int read_adt74x0(const uint8_t addr, double *temp);
static int status_adt74x0(const uint8_t addr);
void exit_on_error(int error_number, int exit_code);

FILE *fp;
double t1, t2, t3, t4;

int main(void)
{
  // Does the lock file exist?
  fp = fopen(LOCK_FILE, "r");
  if (fp != NULL)
    {
    printf("Error 111 - lock file exists, possibly application is already running!\n");
    fclose(fp);
    exit(111);
    }


  // Create lock file
  fp = fopen(LOCK_FILE, "w");
  if (fp == NULL)
    {
    printf("Error 112 - cannot create lock file.\n");
    exit(112);
    }
  fclose(fp);



printf("\n"); fflush(stdout);
int valid1=0; int valid2=0; int valid3=0; int valid4=0;
for(;;)
  {

  // Open I2C bus
  bcm2835_init();
  bcm2835_i2c_begin();

  // slowing down to 10kHz (std is 100kHz) works better
  // when the cables are long and termination dodgy...
  bcm2835_i2c_set_baudrate(10000);

  // Initialize ADT7420 chips & start conversions
  int stat = init_adt74x0(ADT7420_ADDR_1);
  if (stat != 0)
    {
    //exit_on_error(stat,stat);
    }
  bcm2835_delay(1000);  // Allow 1s to settle things


  stat = init_adt74x0(ADT7420_ADDR_2);
  if (stat != 0)
    {
    //exit_on_error(stat,stat);
    }
  bcm2835_delay(1000);  // Allow 1s to settle things

  stat = init_adt74x0(ADT7420_ADDR_3);
  if (stat != 0)
    {
    //exit_on_error(stat,stat);
    }
  bcm2835_delay(1000);  // Allow 1s to settle things

  stat = init_adt74x0(ADT7420_ADDR_4);
  if (stat != 0)
    {
    //exit_on_error(stat,stat);
    }
  bcm2835_delay(1000);  // Allow 1s to settle things


  // Measure temperature
//  printf("\n"); fflush(stdout);
//  int valid1=0; int valid2=0; int valid3=0; int valid4=0;
//  for(;;)
//    {
    // Quit?
    fp = fopen(QUIT_FILE, "r");
    if (fp != NULL)
      {
      fclose(fp);
      break;
      }

    // Get result
    bcm2835_delay(100);  // Avoid race condition

    stat = status_adt74x0(ADT7420_ADDR_1);
    if (stat < 0)
      {
      //exit_on_error(stat,stat);
      }
        else
          {
          if(stat) goto next1;  // No new measurement is ready
          t1;  //Read temperature
          stat = read_adt74x0(ADT7420_ADDR_1, &t1);
          if (stat < 0)
            {
            valid1 = 0;
            //exit_on_error(stat,stat);
            }
              else
                {
                valid1 = 1;
                }
          }
    next1: ;


    stat = status_adt74x0(ADT7420_ADDR_2);
    if (stat < 0)
      {
      //exit_on_error(stat,stat);
      }
        else
          {
          if(stat) goto next2;  // No new measurement is ready
          t2;  //Read temperature
          stat = read_adt74x0(ADT7420_ADDR_2, &t2);
          if (stat < 0)
            {
            valid1 = 0;
            //exit_on_error(stat,stat);
            }
              else
                {
                valid2 = 1;
                }
          }
    next2: ;

    stat = status_adt74x0(ADT7420_ADDR_3);
    if (stat < 0)
      {
      //exit_on_error(stat,stat);
      }
        else
          {
          if(stat) goto next3;  // No new measurement is ready
          t3;  //Read temperature
          stat = read_adt74x0(ADT7420_ADDR_3, &t3);
          if (stat < 0)
            {
            valid1 = 0;
            //exit_on_error(stat,stat);
            }
              else
                {
                valid3 = 1;
                }
          }
    next3: ;

    stat = status_adt74x0(ADT7420_ADDR_4);
    if (stat < 0)
      {
      //exit_on_error(stat,stat);
      }
        else
          {
          if(stat) goto next4;  // No new measurement is ready
          t4;  //Read temperature
          stat = read_adt74x0(ADT7420_ADDR_4, &t4);
          if (stat < 0)
            {
            valid1 = 0;
            //exit_on_error(stat,stat);
            }
              else
                {
                valid4 =1;
                }
          }
    next4: ;



    // Print the outputs
    if (valid1 || valid2 || valid3 || valid4)
      {
      system("./TimeStamp.sh");
      printf("  ");
      if (valid1)
        {
        printf("%.2f  ", t1);
        } else
          {
          printf("-----  ");
          }

     if (valid2)
        {
        printf("%.2f  ", t2);
        } else
          {
          printf("-----  ");
          }

     if (valid3)
       {
       printf("%.2f  ", t3);
       } else
         {
         printf("-----  ");
         }

     if (valid4)
       {
       printf("%.2f  ", t4);
       } else
         {
         printf("-----  ");
         }

    printf("C\n"); fflush(stdout);
    valid1=0; valid2=0; valid3=0; valid4=0;
    }
  bcm2835_i2c_end();
  bcm2835_close();
  //usleep(100000);
  }


  printf("300 C\n");
  bcm2835_i2c_end();
  bcm2835_close();
  unlink(QUIT_FILE);
  unlink(LOCK_FILE);
  return 0;
}



// Return 0 if OK, -ve to show error
static int init_adt74x0(const uint8_t addr)
{
  int stat;
  uint8_t buff[4];

  bcm2835_i2c_setSlaveAddress(addr);

  char reg = IDREG;
  if ((stat = bcm2835_i2c_read_register_rs(&reg, (char *)buff, 1)) != 0)
    return -(0x30 + stat);
  if ((buff[0] & 0xf8) != 0xc8)
    return -0x3f;

  buff[0] = RESET;
  if ((stat = bcm2835_i2c_write((char *)buff, 1)) != 0)
    return -(0x10 + stat);
  bcm2835_delay(1); // Device needs 200us after reset, give it 1ms

  buff[0] = CONFIG;
  buff[1] = ADT7420_CFG;
  if ((stat = bcm2835_i2c_write((char *)buff, 2)) != 0)
    return -(0x20 + stat);

  return 0;
}


// Return 0 if OK, -ve to show error
// Set *temp to be the temperature in Celsius
static int read_adt74x0(const uint8_t addr, double *temp)
{
  int stat;
  uint8_t buff[4];

  bcm2835_i2c_setSlaveAddress(addr);

  char reg = T_MSB;
  if ((stat = bcm2835_i2c_read_register_rs(&reg, (char *)buff, 2)) != 0)
    return -(0x40 + stat);

  // ADT74x0 puts MSB first so flip order
  int16_t hi = buff[0];
  int16_t lo = buff[1];

  int16_t t128 = hi << 8 | lo;

  *temp = t128 / 128.0;

  return 0;
}



// Return 0 or 1 reflecting ADT7420 status register bit 7, -ve to show error
static int status_adt74x0(const uint8_t addr)
{
  int stat;
  uint8_t buff[4];

  bcm2835_i2c_setSlaveAddress(addr);

  char reg = STATUS;
  if ((stat = bcm2835_i2c_read_register_rs(&reg, (char *)buff, 1)) != 0)
    return -(0x40 + stat);

  if ((0x80&buff[0])==0) return 0;
    else return 1;
}


// Close everything and exit on error
void exit_on_error(int error_number, int exit_code)
{
  printf("Error %d\n", error_number);
  bcm2835_i2c_end();
  bcm2835_close();
  unlink(LOCK_FILE);
  exit(exit_code);
}

